"""
 * Author:    Jonas Gjendem Røysland
 * Created:   14.06.2023
"""


import copy 
import matplotlib.pyplot as plt
import struct
import numpy as np
import sys 
import os

def read_compressed_labels(file_path):
     
     #Functions that analyse the binary file generated by the SVMBDT 
     #and converts it to a readable execution [int], n_classes[], and 
     #the predicted labels[].
     
    with open(file_path, mode='rb') as file: # b is important -> binary
        fileContent = file.read()
    filesize = len(fileContent)
    unpack_opt = "ii16B" + str(filesize-4-4-16) + "B"
    arr = struct.unpack(unpack_opt, fileContent)

    classification_execution_time = arr[0]
    loading_execution_time = arr[1]
    classes_holder = arr[2:18]
    labels_holder = arr[18:]

    classes = []
    labels = []

    
    # Decode the labels and converts 
    # them back to original classes. 

    for i in range(len(classes_holder)): 
        if(int(classes_holder[i]) != 255): 
            classes.append(classes_holder[i])

    if(len(classes) <= 2): 
        for i in range(len(labels_holder)): 
            pixel_str = format(labels_holder[i], "b").zfill(8)
            for j in range(8): 
                labels.append(int(pixel_str[j],2))

    if(len(classes) > 2 and len(classes) <= 4): 
        for i in range(len(labels_holder)): 
            pixel_str = format(labels_holder[i], "b").zfill(8)
            for j in range(4): 
                labels.append(int(pixel_str[2*j:(j+1)*2],2))

    if(len(classes) > 4 and len(classes) <= 16):
        for i in range(len(labels_holder)): 
            pixel_str = format(labels_holder[i], "b").zfill(8)
            for j in range(2): 
                labels.append(int(pixel_str[4*j:(j+1)*4],2)) 

    for i in range(len(labels)): 
        labels[i] = classes[labels[i]]

    return classification_execution_time, loading_execution_time, classes, labels


def rotate_labels(labeled_image, x_dim, y_dim):
    
    resized_li = np.array(labeled_image)
    resized_li.resize(y_dim,x_dim)
    rotate_image = []

    for i in range(len(resized_li[0])): 
        for j in range(len(resized_li)): 
            rotate_image.append(resized_li[j][i])
    return np.array(rotate_image).flatten()


def reduce_and_rotate_pixels(labels,y_dim): 
        new_label = []
        for i in range(0,len(labels),y_dim*3): 
            new_label.append(labels[i:i+y_dim])

        new_label = np.array(new_label).flatten()
    
        return new_label

def store_labeled_image(predicted_label, x_dim, y_dim, path): 

    def color_labels(labels, dimx, dimy, color_label): 

        
        #Define a new color scheme for the labels. 
        #Each classes are defined with the RGB code in hex, 0-255.
        #Color labeling is defined by: color_label = {0 : [r,g,b], ..., N : [r,g,b]}, 
        #with N amount of classes. 

        #Returns the new_new_label with a new color scheme. 
        

        new_label = copy.deepcopy(labels)
        new_label = new_label.reshape(dimx, dimy)
        new_new_label = []

        for i in range(len(new_label)): 
            x_array = []
            for j in range(len(new_label[0])): 
                x_array.append(color_label[new_label[i][j]])
            new_new_label.append(x_array)
        
        return np.array(new_new_label)
    
    #Color labels in RGB. 
    hypso_color_label = { 0 : [0,0,139], 
                            1 : [13,152,186], 
                            2 : [144,238,144], 
                            3 : [20,70,50], 
                            4 : [240,0,0], 
                            5 : [107, 74, 58], 
                            6 : [255,255,255], 
                            7 : [194,178,128], 
                            8 : [180,180,180], 
                            9 : [100,100,100],  
                            10 : [0,0,0]
    }

    new_predicted_label = copy.deepcopy(predicted_label)
    new_predicted_label = np.array(new_predicted_label)

    np.save(path, new_predicted_label)

    yout_color = color_labels(new_predicted_label,x_dim,y_dim, hypso_color_label)

    plt.imshow(yout_color)
    plt.axis('off')
    plt.savefig(path, dpi = 300, bbox_inches="tight", pad_inches = 0)
    plt.close()




def main():     
    
    #    - The python script produces two files to the desired path
    #        - <decoded labels save path>.npy
    #        - <decoded labels save path>.png

        hsi_default_dimension = [[684,956],[1092,598],[1216,537]] #Add more dimension here if the need for it.
        
        check_for_help = ["-h", "-help"]

        try: 
            compressed_labels_path = str(sys.argv[1])
        except: 
            print("How-to-run with default the default dimensions: ", str(hsi_default_dimension).replace("[", "(").replace("]",")"))
            print("python3 decode_labels.py <compressed labels bin file path> <decoded labels save path>")
            print("")
            print("How-to-run with custom dimensions, print metadata and skew mode:")
            print("python3 decode_labels.py <compressed labels bin file path> <decoded labels save path> <x dim> <y dim> <print metadata> <skew mode>")
            return 
            
        if(sys.argv[1] in check_for_help): 
            print("How-to-run with default the default dimensions: ", str(hsi_default_dimension).replace("[", "(").replace("]",")"))
            print("python3 decode_labels.py <compressed labels bin file path> <decoded labels save path>")
            print("")
            print("How-to-run with custom dimensions, print metadata and skew mode:")
            print("python3 decode_labels.py <compressed labels bin file path> <decoded labels save path> <x dim> <y dim> <print metadata> <skew mode>")
            return 
        

        if(not(os.path.isfile(compressed_labels_path))): 
            print("Labels file path is incorrect or doesn't exists!")
            return 

        classification_execution_time, loading_execution_time, classes, labels = read_compressed_labels(compressed_labels_path)
        labels_save_path = str(sys.argv[2])

        check_for_arguments = 0
        
        try: 
            if(sys.argv[3] and sys.argv[4] and 
               sys.argv[5] and sys.argv[6]): 
                check_for_arguments = 1
        except: 
            None

        if(check_for_arguments == 0):
            x_dim = 0
            y_dim = 0

            for i in range(len(hsi_default_dimension)): 
                if(len(labels)%hsi_default_dimension[i][0] == 0 and len(labels)%hsi_default_dimension[i][1] == 0): 
                    x_dim = hsi_default_dimension[i][0]
                    y_dim = hsi_default_dimension[i][1]
                    break 
            
            if(x_dim == 0 and y_dim == 0): 
                print("The labels didn't correponds to the dimensions!")
                print("Size of labels: ", len(labels))
                print("")
                print("Length of the default dimensions: ")
                for i in range(len(hsi_default_dimension)): 
                    print(str(hsi_default_dimension[i][0]) +
                           "x" + str(hsi_default_dimension[i][1]) + "=" + hsi_default_dimension[i][0] * hsi_default_dimension[i][1])
                return
            try:            
                labels = rotate_labels(labels,x_dim, y_dim)
                labels = reduce_and_rotate_pixels(labels, y_dim)
                if(x_dim == 1216 and y_dim == 537): 
                    x_dim = int(x_dim/3)+1
                    labels = labels.reshape(x_dim, y_dim)
                    labels = np.flip(labels, axis=0).flatten()
                else: 
                    x_dim = int(x_dim/3)
                store_labeled_image(labels, x_dim, y_dim, labels_save_path)
                return 0
            except: 
                print("Default dimension dimensions didn't correspond with labels")
                print("Test with customs dimensions and skew==0")

        if(check_for_arguments == 1): 
            x_dim = int(sys.argv[3])
            y_dim = int(sys.argv[4])
            print_execution_time = int(sys.argv[5])
            skew = int(sys.argv[6])
            
            if((len(labels)%x_dim == 0) and (len(labels)%y_dim == 0)): 
                if(skew == 0):
                    store_labeled_image(labels, y_dim, x_dim, labels_save_path)
                if(skew == 1):
                    try: 
                        labels = rotate_labels(labels,x_dim, y_dim)
                        labels = reduce_and_rotate_pixels(labels, y_dim)
                        x_dim = int(x_dim/3)
                        store_labeled_image(labels, x_dim, y_dim, labels_save_path)
                    except: 
                        print("Dimensions with skew didn't correspond!")
                        print("Try change <x dim> and <y dim> with each other!")
                        return
            else: 
                print("The labels didn't correponds to the dimensions!")
                print("Size of labels: ", len(labels))
                print("Size of typed dimensions: x dim * y dim =", x_dim*y_dim)
                return

            if(print_execution_time == 1):
                print("Classes in labels:", classes) 
                print("Classification Execution Time:", round(classification_execution_time/1000000,2), "s")
                print("Loading Execution Time:", round(loading_execution_time/1000,2), "ms")
            
            return 0

main()
