/*
 * Author:    Jonas Gjendem Røysland
 * Created:   14.06.2023
*/

#include "../inc/onboard_classification.h"

uint8_t onboard_classification_predict(
    char capture_path[MAX_CHAR_PATH], 
    char save_labels_path[MAX_CHAR_PATH],
    char config_folder_path[MAX_CHAR_PATH], 
    uint8_t processing_mode, 
    uint8_t bdt_index_number, 
    char local_angels_path[MAX_CHAR_PATH]){

    //Load bip cube 

    FILE* fd;

    char binned_bip_path[MAX_CHAR_PATH];
    strcpy(binned_bip_path, capture_path); 
    strcat(binned_bip_path, CUBE_BIP_NAME); 

    fd = fopen(binned_bip_path, "rb"); 

    if(!fd){
        DEBUG(printf("Onboard Classification: Loading binned bip file failed! \n");)
        return -1; 
    }
 
    else{
        //Find File Size 

        struct stat st;
        stat(binned_bip_path, &st);
        uint32_t n_pixels = (uint32_t)(st.st_size/(SPECTRAL_BAND_SIZE*sizeof(uint16_t)));
    
        if(n_pixels <= MAX_HYPSO_CUBE_SIZE){
                //Allocate Memory to Structs
            struct hsi_cube_pixel *cube_pixel = 
                (struct hsi_cube_pixel*)malloc(sizeof(struct hsi_cube_pixel));

            struct calibrated_cube_pixel *calibrated_pixel = 
                (struct calibrated_cube_pixel*)malloc(sizeof(struct calibrated_cube_pixel));

            struct Spectral_calibration *sc = 
                (struct Spectral_calibration*)malloc(sizeof(struct Spectral_calibration));
            
            struct radiometric_calibration_coeff *rcc = 
                (struct radiometric_calibration_coeff*)malloc(sizeof(struct radiometric_calibration_coeff));
                
            struct extraterrestrial_solar_irradiance *esi = 
                (struct extraterrestrial_solar_irradiance*)malloc(sizeof(struct extraterrestrial_solar_irradiance));

            struct local_angels *la = 
                (struct local_angels*)malloc(sizeof(struct local_angels));

            struct Labeled_Image *image = 
                (struct Labeled_Image*)malloc(sizeof(struct Labeled_Image));

            float radiometric_metadata[3]; 
            uint8_t svmbdt_metadata[1]; 

            uint8_t bdt_model[MAX_CLASSES][BDT_INDEX_LENGTH];
            float svm_models[MAX_CLASSES][SPECTRAL_BAND_SIZE+10]; 
            uint8_t spectral_band_list[SPECTRAL_BAND_SIZE];

            struct timeval stop, start;
            gettimeofday(&start, NULL);
            
            int8_t ret_radiometric = 0; 
            int8_t ret_svmbdt = 0; 

            //Loading config values
            ret_radiometric = radiometric_calibration_init(
                capture_path, config_folder_path, local_angels_path, 
                processing_mode, radiometric_metadata, sc, rcc, esi, la); 

            ret_svmbdt = svmbdt_init(
                config_folder_path, bdt_index_number, processing_mode, 
                bdt_model, svm_models, spectral_band_list, svmbdt_metadata); 

            gettimeofday(&stop, NULL);

            uint32_t loading_execution_time = (stop.tv_sec - start.tv_sec) * 1000000 
                                + stop.tv_usec - start.tv_usec;

            DEBUG(printf("Loading Config data execution time %i us \n", loading_execution_time);)
            image->loading_execution_time = loading_execution_time;  

            float exposure = radiometric_metadata[0]; 
            float background_value = radiometric_metadata[1]; 
            float earth_sun_distance = radiometric_metadata[2];

            uint8_t n_bands_used = svmbdt_metadata[0]; 
            
            uint16_t buffer[SPECTRAL_BAND_SIZE];
            uint16_t radio_index = 0; 
            uint16_t local_angels_index = 0; 

            gettimeofday(&start, NULL);

            if(ret_radiometric >= 0 && ret_svmbdt >= 0){

                uint16_t radio_index_max = 0; 
                uint16_t local_angels_max = 0; 

                if(n_pixels%CUBE_Y_LENGTH == 0){
                    radio_index_max = CUBE_Y_LENGTH; 
                }

                if(n_pixels%CUBE_X_LENGTH == 0){
                    local_angels_max = CUBE_X_LENGTH;
                }

                //Starting Calibration and Prediction Loop. 
                //Reading pixelwise from cube to use less memory.
                for(uint32_t pixel_index = 0; pixel_index < n_pixels; pixel_index++){

                    fread(buffer, sizeof(uint16_t), SPECTRAL_BAND_SIZE, fd); 

                    for(uint16_t j = 0; j < SPECTRAL_BAND_SIZE; j++){
                        cube_pixel->X[j] = buffer[SPECTRAL_BAND_SIZE-j-1];
                    }

                    //Calibrate pixel
                    radiometric_calibration_pixel(
                        cube_pixel, calibrated_pixel, processing_mode, 
                        exposure, sc, rcc, esi, la, background_value, 
                        earth_sun_distance, radio_index, local_angels_index); 

                    radio_index++; 
                    if(radio_index >= radio_index_max){
                        radio_index = 0; 
                    }

                    local_angels_index++; 
                    if(local_angels_index >= local_angels_max){
                        local_angels_index = 0; 
                    }
                    
                    //Predict pixel
                    svmbdt_predict(calibrated_pixel->X, 
                        bdt_model, image, svm_models, spectral_band_list, 
                        n_bands_used, processing_mode, pixel_index);
                }

                gettimeofday(&stop, NULL);
                    
                uint32_t classification_execution_time = (stop.tv_sec - start.tv_sec) * 1000000 
                                    + stop.tv_usec - start.tv_usec;

                DEBUG(printf("Classification execution time %i us \n", classification_execution_time);) 

                image->classification_execution_time = classification_execution_time; 

                //Save Labeled Image
                labeled_image_bin(save_labels_path, image, n_pixels); 

            }

            // Free up allocated space
            free(cube_pixel); 
            free(calibrated_pixel); 
            free(sc);
            free(rcc); 
            free(esi);
            free(la);
            free(image);

        }

        fclose(fd);

        return 0; 
    }
}


uint8_t onboard_classification_execute(uint8_t argc, char *argv[]){

    if(argc >= 6){ 

        char capture_path[MAX_CHAR_PATH] = "";
        char save_labels_path[MAX_CHAR_PATH] = "";
        char config_file_path[MAX_CHAR_PATH] = ""; 
        char bdt_number_char[MAX_CHAR_PATH] = ""; 
        char processing_mode_char[MAX_CHAR_PATH] = ""; 
        char local_angels_path[MAX_CHAR_PATH] = ""; 

        uint8_t processing_mode = 0; 
        uint8_t bdt_number = 1; 

        strcpy(capture_path,argv[1]);
        strcpy(save_labels_path,argv[2]);
        strcpy(config_file_path, argv[3]); 
        strcpy(bdt_number_char, argv[4]);
        strcpy(processing_mode_char, argv[5]);

        processing_mode = strtoul(processing_mode_char,0,10); 
        bdt_number = strtoul(bdt_number_char,0,10); 
                
        if(processing_mode == 2 || processing_mode == 3){ 
            if(argc == 7){
                strcpy(local_angels_path, argv[6]);
            }
            else{
                DEBUG(printf("Onboard Classification: Local Angels Path not defined!\n"));
                return -1; 
            }
        }


        onboard_classification_predict(
            capture_path, 
            save_labels_path, 
            config_file_path, 
            processing_mode, 
            bdt_number, 
            local_angels_path);


    }
    
    else{
        DEBUG(printf("Onboard Classification: Too few arguments! \n"));
        return -1; 
    }

    return 0; 

}


int main(int argc, char *argv[]){

    onboard_classification_execute(argc, argv); 

    return 0; 
}


/*
0 radiance
1 radiance_sparse 
2 reflectence 
3 reflectence_sparse 
*/
