/*
 * Author:    Jonas Gjendem Røysland
 * Created:   14.06.2023
*/


#include "../inc/radiometric.h"


// Private variabels 
static uint32_t hsi_pixels; 

// Private Functions

float degress_to_radians(float deg_num){
    return deg_num * (PI/180.0);
}


int8_t load_extraterrestrial_solar_irradiance(char config_folder_path[MAX_CHAR_PATH], 
    struct extraterrestrial_solar_irradiance *esi){

    char esi_path[MAX_CHAR_PATH] = "";
    strcpy(esi_path, config_folder_path); 
    strcat(esi_path, IRRADIANCE_THUILLIER_PATH); //Can be changed for CHKUR.

    FILE* fd;

    fd = fopen(esi_path, "rb"); 

    if(!fd){
        DEBUG(printf("Radiometric: Loading extraterrestrial solar irradiance failed! \n");)
        return -1; 
    }
    else{
        int8_t ret = fread(esi->coeff, sizeof(float), SPECTRAL_BAND_SIZE, fd); 
        fclose(fd);
        return 0; 

    }
}


int8_t load_local_angels(char path[MAX_CHAR_PATH], 
            struct local_angels *la){

    FILE* fd;

    fd = fopen(path, "r"); 

    if(!fd){
        DEBUG(printf("Radiometric: Loading Local angels failed! \n");)
        return -1; 
    }

    else{
        char buffer[4096]; 
        char *token, *line; 
        uint16_t i = 0; 
        uint16_t j; 

        line=fgets(buffer,sizeof(buffer),fd);
    
        while((line=fgets(buffer,sizeof(buffer),fd))!=NULL){
        token = strtok(line,",");
        j = 0;
        while(token != NULL){
            if(i < CUBE_X_LENGTH && j < 8){
                la->coeff[i][j++] = atof(token);
            }
            token = strtok(NULL,",");
            }
        ++i ;
        }

        fclose(fd); 
        return 0; 
    }
}



int8_t get_capture_config_file_path(char capture_path[MAX_CHAR_PATH], 
    char capture_config_path[MAX_CHAR_PATH]){

    char default_capture_config_path[MAX_CHAR_PATH] = CAPTURE_CONFIG_NAME;
    char config_path_holder[MAX_CHAR_PATH];
    char capture_path_holder[MAX_CHAR_PATH];

    strcpy(capture_path_holder, capture_path); 

    char * token = strtok(capture_path_holder, "/");
    while(token != NULL){
        if(token != NULL){
            strcpy(config_path_holder, token);
        }
        token = strtok(NULL, "/");
    }

    strcpy(capture_config_path, capture_path);
    strcat(capture_config_path, "/");
    strcat(capture_config_path, config_path_holder);
    strcat(capture_config_path, default_capture_config_path);

    return 0; 

}


float find_exposure_ini(char capture_path[200]){

    char exposure_path[MAX_CHAR_PATH] = ""; 
    strcpy(exposure_path, capture_path); 
    strcat(exposure_path, CAPTURE_CONFIG_NAME); 

    FILE* fd;

    fd = fopen(exposure_path, "r"); 

    if(!fd){
        DEBUG(printf("Radiometric: Loading exposure ini failed! \n");)
        return -1; 
    }

    else{
        float exposure = 0; 
        char buffer[100]; 
        char * token; 
        char exposure_char[10] = "exposure"; 


        while(fgets(buffer, 100, fd)){
            if(strstr(buffer, exposure_char) != NULL){
                token = strtok(buffer, "exposure = ");
                exposure = strtof(token,NULL);
            }
        }

        fclose(fd);

        return exposure * 0.001; 
    }
}


int8_t read_spectral_calibration(char path[MAX_CHAR_PATH], 
    struct Spectral_calibration *sc){

    FILE* fd;

    fd = fopen(path, "r"); 

    if(!fd){
        DEBUG(printf("Radiometric: Loading Spectral calibration failed! \n");)
        return -1; 
    }
    
    else{

        for(uint8_t i = 0; i < SPECTRAL_CALIBRATION_LENGTH; i++){
            sc->spectral_calibration[i] = 0; 
        }

        char buffer[30]; 

        for(uint8_t i = 0; i < SPECTRAL_CALIBRATION_LENGTH; i++){
            fgets(buffer, 30, fd);
            sc->spectral_calibration[i] = strtof(buffer,NULL);
        }

        fclose(fd); 

        return 0; 
    }
}

int8_t read_radiometric_calibration_coeff(char path[MAX_CHAR_PATH], 
            struct radiometric_calibration_coeff *rcc){

    FILE* fd;

    fd = fopen(path, "r"); 

    if(!fd){
        DEBUG(printf("Radiometric: Loading Radiometric calibration coeff failed! \n");)
        return -1; 
    }

    else{

        char buffer[4096]; 
        char *token, *line; 
        uint16_t i = 0; 
        uint16_t j; 
    
        while((line=fgets(buffer,sizeof(buffer),fd))!=NULL){
        token = strtok(line,",");
        j = 0;
        while(token != NULL){
            if(i < CUBE_Y_LENGTH && j < SPECTRAL_BAND_SIZE){
                rcc->coeff[i][j++] = atof(token);
            }
            token = strtok(NULL,",");
            }
        ++i ;
        }

        fclose(fd); 

        return 0; 
    }
}



int8_t read_dark_frame(char path[MAX_CHAR_PATH], 
            struct dark_frame *df){

    FILE* fd;

    fd = fopen(path, "r"); 

    if(!fd){
        DEBUG(printf("Radiometric: Loading Dark Frame failed! \n");)
        return -1; 
    }

    else{

        char buffer[DARK_FRAME_BUFFER_SIZE]; 
        char *token, *line; 
        uint16_t i = 0; 
        uint16_t j; 

        while((line=fgets(buffer,sizeof(buffer),fd))!=NULL){
        token = strtok(line,",");
        j = 0;
        while(token != NULL){
            if(i < CUBE_Y_LENGTH && j < DARK_FRAME_LENGTH){
                df->frame[i][j++] = atof(token); //Segmentation Fault
            }
            token = strtok(NULL,",");
        }
      
        ++i ;
      
        }

        fclose(fd); 

        return 0; 
    }
}

float average(struct dark_frame *df){ // Might be replaced as a macro. Should be!!!
    double average_holder = 0; 
    
    for(uint16_t i = 0; i < CUBE_Y_LENGTH; i++){
        for(uint16_t j = 0; j < DARK_FRAME_LENGTH; j++){
            average_holder += df->frame[i][j];
        }
    }

    return (float)(average_holder)/(CUBE_Y_LENGTH*DARK_FRAME_LENGTH);
}

int8_t load_calibration_files(
        char config_folder_path[MAX_CHAR_PATH], 
        struct Spectral_calibration *sc,
        struct radiometric_calibration_coeff *rcc){

    // Load spectral calibrations
    char spectral_calibration_path[MAX_CHAR_PATH] = "";
    char radiometric_coeff_path[MAX_CHAR_PATH] = "";
    char dark_path[MAX_CHAR_PATH] = "";
    
    strcpy(spectral_calibration_path, config_folder_path); 
    strcat(spectral_calibration_path, SPECTRAL_CALIBRATION_PATH); 

    strcpy(radiometric_coeff_path, config_folder_path);
    strcat(radiometric_coeff_path, RADIOMETRIC_COEF_PATH); 

    //strcpy(dark_path, config_folder_path); 
    //strcat(dark_path, DARK_PATH);

    int8_t ret_spectral = 0;
    int8_t ret_radiometric = 0;  

    ret_spectral = read_spectral_calibration(spectral_calibration_path, sc);
    ret_radiometric = read_radiometric_calibration_coeff(radiometric_coeff_path,rcc);
    //ret_t = read_dark_frame(dark_path, df);

    if(ret_radiometric < 0 && ret_spectral < 0){
        return -1; 
    }
    
    else{
        return 0; 
    }
}


int8_t linspace(float start, 
    float stop, 
    float n_steps, 
    float arr[SPECTRAL_BAND_SIZE]){
    
    float step = (stop - start)/n_steps;
    step = (stop - start + step)/n_steps;
    float num  = start; 

    for(uint16_t i = 0; i < n_steps; i++){
        arr[i] = num;
        num = num + step;  
    }

    return 0;

}

int8_t pixel_to_wavelength(float x[SPECTRAL_BAND_SIZE], 
    float spectral_coeff[SPECTRAL_CALIBRATION_LENGTH], 
    float w[SPECTRAL_BAND_SIZE]){

    //Check length of spectral_coeff
    uint8_t n_coeff = 0; 
    for(uint8_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
        if(spectral_coeff[i] != 0){
            n_coeff++; 
        }
    }

    if(n_coeff == 2){
        for(uint8_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
            w[i] = spectral_coeff[1] + spectral_coeff[0] * x[i];
        }

    }
    
    if(n_coeff == 3){
        for(uint8_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
            w[i] = spectral_coeff[2] + spectral_coeff[1] * x[i] + 
            spectral_coeff[0] * x[i] * x[i];
        }
        
    }

    if(n_coeff == 4){
        for(uint8_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
            w[i] = spectral_coeff[3] + spectral_coeff[2] * x[i] + 
            spectral_coeff[1] * x[i] * x[i] + spectral_coeff[0] * x[i] * x[i] * x[i];
        }
        
    }

    if(n_coeff == 5){
        for(uint8_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
            w[i] = spectral_coeff[4] + spectral_coeff[3] * x[i] + spectral_coeff[2] * x[i] * x[i] + 
            spectral_coeff[1] * x[i] * x[i] * x[i] + spectral_coeff[0] * x[i] * x[i] * x[i] * x[i];
        }
        
    }

    return 0; 
}


int8_t apply_spectral_calibration(uint16_t x_start, uint16_t x_stop, 
    uint32_t image_width,
    struct Spectral_calibration *sc,
    float w[SPECTRAL_BAND_SIZE]){
    
    float x[SPECTRAL_BAND_SIZE]; 
    linspace(x_start, x_stop, image_width, x);
    pixel_to_wavelength(x,sc->spectral_calibration,w);

    return 0; 
}

int8_t apply_radiometric_calibration(float pixel[SPECTRAL_BAND_SIZE], 
        float exp,
        float background_value, 
        float radiometric_calibration_coeff[SPECTRAL_BAND_SIZE]){

    if(exp < 0.001){
        for(uint32_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
            pixel[i] = pixel[i] - background_value;
            pixel[i] = pixel[i] * radiometric_calibration_coeff[i] / (0.01);
         }
    }

    else{
        for(uint32_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
            pixel[i] = pixel[i] - background_value;
            pixel[i] = pixel[i] * radiometric_calibration_coeff[i] / exp;
        }
    }

    return 0; 
}


int8_t calibrate_cube_pixel(struct hsi_cube_pixel *cube, 
    float metadata[5], 
    struct Spectral_calibration *sc, 
    struct radiometric_calibration_coeff *rcc, 
    float background_value,
    uint32_t radio_index, 
    struct calibrated_cube_pixel *radiance_cube_pixel){

    //float w[SPECTRAL_BAND_SIZE];
    //apply_spectral_calibration(metadata[3], metadata[4], metadata[2], sc, w); 

    for(uint8_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
        radiance_cube_pixel->X[i] = cube->X[i];
    }

    apply_radiometric_calibration(radiance_cube_pixel->X, metadata[0], background_value, rcc->coeff[radio_index]);

    return 0; 
}


int8_t calibrate_hypso_cube(
    struct hsi_cube_pixel *cube_pixel, 
    struct calibrated_cube_pixel *calibrated_pixel, 
    float exposure, 
    struct Spectral_calibration *sc, 
    struct radiometric_calibration_coeff *rcc, 
    float background_value,
    uint32_t radio_index){
    
    //float background_value = average(df); 
    float x_start = 428;  // aoi_x
    float x_stop = 1508;  // aoi_x + column_count
    float image_height = 684;  // row_count
    float image_width = 120;  // column_count/bin_factor

    float metadata[5] = {exposure, image_height, image_width, x_start, x_stop};

    calibrate_cube_pixel(cube_pixel, metadata, sc, rcc, background_value, radio_index, calibrated_pixel);

    return 0; 
}

uint16_t get_day_of_year(time_t timestamp){
    struct tm *ts;
    ts = localtime(&timestamp);
    return ts->tm_yday;
    
}


float calculate_earth_sun_distance(time_t timestamp){

    uint16_t day_number = get_day_of_year(timestamp); 

    float earth_sun_distance_table[DISTANCE_TABLE_SIZE][2] = 
        {{1, 0.98331}, {15, 0.98365}, {32, .98536}, {46, 0.98774}, {60, 0.99084},
        {74, 0.99446}, {91, 0.99926}, {106, 1.00353}, {121, 1.00756}, {135, 1.01087},
        {152, 1.01403}, {166, 1.01577}, {182, 1.01667}, {196, 1.01646}, {213, 1.01497},
        {227, 1.01281}, {242, 1.00969}, {258, 1.00566}, {274, 1.00119}, {274, 1.00119},
        {288, 0.99718}, {305, 0.99253}, {319, 0.98916}, {335, 0.98608}, {349, 0.98426},
        {365, 0.98333}};

    uint8_t index_day = 0; 

    for(uint8_t i = 0; i < DISTANCE_TABLE_SIZE; i++){
        if(earth_sun_distance_table[i][0] - day_number >= 0){
            index_day = i; 
            break; 
        }
    }

    if(index_day >= DISTANCE_TABLE_SIZE || index_day < 0){
        index_day = 13; 
    }


    return earth_sun_distance_table[index_day][1];

}

int8_t radiance_to_reflectance(
    struct calibrated_cube_pixel *calibrated_pixel, 
    float earth_sun_distance, 
    struct local_angels *la,
    struct extraterrestrial_solar_irradiance *esi, 
    uint16_t local_angels_index){


    for(uint16_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
        calibrated_pixel->X[i] = 
        calibrated_pixel->X[i] * 
        PI * 
        earth_sun_distance * earth_sun_distance *  //Squared ^2
        esi->coeff[i] * 
        la->coeff[local_angels_index][SOLAR_ZENITH_ANGLE_INDEX];
        
        }

        return 0; 
}


int8_t radiometric_calibration_init(
    char capture_path[MAX_CHAR_PATH], 
    char config_folder_path[MAX_CHAR_PATH], 
    char local_angels_path[MAX_CHAR_PATH],
    uint16_t processing_mode, 
    float radiometric_metadata[3], 
    struct Spectral_calibration *sc, 
    struct radiometric_calibration_coeff *rcc, 
    struct extraterrestrial_solar_irradiance *esi, 
    struct local_angels *la){

    int8_t ret_calibration = 0; 
    int8_t ret_exp = 0; 

    ret_calibration = load_calibration_files(config_folder_path,sc,rcc);
    radiometric_metadata[0] = find_exposure_ini(capture_path);

    if(radiometric_metadata[0] < 0){
        ret_exp = -1; 
    }

    radiometric_metadata[1] = DARK_FRAME_BACKGROUND_VALUE; 


    struct timeval timestamp;
    gettimeofday(&timestamp, NULL);
    time_t timestamp_sec = (time_t)timestamp.tv_sec; 
    radiometric_metadata[2] = calculate_earth_sun_distance(timestamp_sec);


    if(ret_calibration >= 0 && ret_exp >= 0){        
        if(processing_mode == 2 || processing_mode == 3){

            int8_t ret_angels = 0; 
            int8_t ret_solar = 0; 

            ret_angels = load_local_angels(
                local_angels_path, la);
            ret_solar = load_extraterrestrial_solar_irradiance(
                config_folder_path,esi);

            if(ret_angels >= 0 && ret_solar >= 0){
                for(uint32_t i = 0; i < CUBE_X_LENGTH; i++){
                    la->coeff[i][SOLAR_ZENITH_ANGLE_INDEX] = 
                    (1/cos(degress_to_radians(la->coeff[i][SOLAR_ZENITH_ANGLE_INDEX])));
                }

                for(uint16_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
                    esi->coeff[i] = (1/esi->coeff[i]);
                }
                return 0; 
            }

            else{
                DEBUG(printf("Radiometric: Loading Config Data Failed! \n");)
                return -1; 
            }
        }
        return 0; 
    }
  
    else{
        DEBUG(printf("Radiometric: Loading Exposure Data Failed! \n");)
        return -1; 
    }

}


int8_t radiometric_calibration_pixel(
    struct hsi_cube_pixel *cube_pixel, 
    struct calibrated_cube_pixel *calibrated_pixel, 
    uint16_t processing_mode, 
    float exposure, 
    struct Spectral_calibration *sc, 
    struct radiometric_calibration_coeff *rcc, 
    struct extraterrestrial_solar_irradiance *esi, 
    struct local_angels *la,
    float background_value, 
    float earth_sun_distance, 
    uint32_t radio_index, 
    uint16_t local_angels_index){


    calibrate_hypso_cube(
        cube_pixel, calibrated_pixel, exposure, sc,rcc,background_value, radio_index); 

    if(processing_mode == 2 || processing_mode == 3){
        radiance_to_reflectance(
            calibrated_pixel, earth_sun_distance, la, esi, local_angels_index);
    }

    return 0; 
}


/*
0 radiance
1 radiance_sparse 
2 reflectence 
3 reflectence_sparse 
*/
